import os
from typing_extensions import Union, Callable
import jax
import jax.numpy as jnp
from scipy import constants

from . import units, trace_utils

prmpar_map = {14: "p", 5626: "Fe"}


def get_arrival_times(
    traces: jax.typing.ArrayLike,
    trace_times: jax.typing.ArrayLike,
) -> jax.typing.ArrayLike:
    """
    Get the arrival times from the trace for each antenna.

    To determine this, we apply a Hilbert envelope and get the maximum position.

    Parameter
    ---------
    traces : jax.typing.ArrayLike
        the traces to take the Hilbert envelope of. Shape must be in
        (Npol, Nant, Nsamples)
    trace_times L jax.typing.ArrayLike
        the time array of the traces. Must be in shape (Nant, Nsamples)
    """
    traces_hilbert = trace_utils.hilbert(traces)
    hilbert_envelope = jnp.sqrt(
        jnp.linalg.norm(traces, axis=0) ** 2
        + jnp.linalg.norm(traces_hilbert, axis=0) ** 2
    )
    arrtime_idces = jnp.argmax(jnp.abs(hilbert_envelope), axis=1)

    arr_times_ant = []
    for iant, antmax_idx in enumerate(arrtime_idces):
        arr_times_ant.append(trace_times[iant, antmax_idx])

    return jnp.array(arr_times_ant)


def gaisser_hillas_function_coreas(
    X: jax.typing.ArrayLike,
    nmax: float,
    x0: float,
    xmax: float,
    p0: float,
    p1: float,
    p2: float,
) -> jax.Array:
    """Return a Gaisser-Hillas function."""
    lmbda = p0 + p1 * X + p2 * X**2  # interaction length, parameterised by polynomial
    return (
        nmax
        * ((X - x0) / (xmax - x0)) ** ((xmax - x0) / lmbda)
        * jnp.exp((xmax - X) / lmbda)
    )


def gaisser_hillas_function(
    x: jax.typing.ArrayLike, nmax: float, xmax: float, x0: float, lmbda: float
) -> jax.Array:
    """Return a Gaisser-Hillas function in standard formalism."""
    return (
        nmax
        * ((x - x0) / (xmax - x0)) ** ((xmax - x0) / lmbda)
        * jnp.exp((xmax - x) / lmbda)
    )


def gaisser_hillas_function_error(
    x: jax.typing.ArrayLike,
    nmax_tup: tuple,
    xmax_tup: tuple,
    x0_tup: tuple,
    lmbda_tup: tuple,
) -> jax.Array:
    """Return the error of the Gaisser Hillas function, using mean and std of GH parameters."""
    nmax, nmax_err = nmax_tup
    xmax, xmax_err = xmax_tup
    x0, x0_err = x0_tup
    lmbda, lmbda_err = lmbda_tup

    gh = gaisser_hillas_function(x, nmax, xmax, x0, lmbda)
    ln_term = jnp.log(((x - x0) / (xmax - x0)))

    x0_err_term = ln_term * x0 - x * ln_term - (xmax - x) / (lmbda * (x0 - x))
    lmbda_err_term = ((xmax - x0) - ln_term + (xmax - x)) / (lmbda**2)

    return jnp.sqrt(
        (gh / nmax * nmax_err) ** 2
        + (gh * ln_term / lmbda * xmax_err) ** 2
        + (gh * x0_err_term * x0_err) ** 2
        + (gh * lmbda_err_term * lmbda_err) ** 2
    )


def gaisser_hillas_function_LR(
    x: jax.typing.ArrayLike, nmax: float, xmax: float, L: float, R: float
) -> jax.Array:
    """Return a Gaisser-Hillas function in L-R formalism"""
    return nmax * jnp.exp((xmax - x) / (L * R)) * (1 + R * (x - xmax) / L) ** (R**-2)


def convert_GH_to_LR(xmax, x0, lmbda) -> tuple[float, float]:
    """
    Convert the standard Gaisser-Hillas parameters to those in the LR formalism.

    Parameter:
    ---------
    xmax : float
        the atmospheric depth at shower maximum
    x0 : float
        the first interaction point in the shower
    lmbda : float
        the interaction length
    Return:
    ------
    L, R : tuple[float, float]
        the width and asymmetry parameter of the profile
    """
    x0p = xmax - x0
    return (jnp.sqrt(jnp.abs(x0p * lmbda)), jnp.sqrt(jnp.abs(lmbda / x0p)))


def get_time_axes(
    trace: jax.typing.ArrayLike,
    arrival_times: jax.typing.ArrayLike,
    delta_t: float = 2 * units.ns,
    sample_axis: int = 2,
) -> jax.Array:
    """
    Get the time axes for traces with all antennas.

    Parameter
    -----------
    traces : jax.typing.ArrayLike
        the trace in which to take the time axes of.
        shape of (Nant, Nsamples)
    arrival_times : jax.typing.ArrayLike
        the arrival time for all antennas. Shape of Nant
    delta_t : float, default=2ns
        the timing resolution in internal units
    sample_axis : int, defauilt=2
        the axis in which the samples are in the trace

    Return
    ------
    trace_times : jax.Array
        the time axes of the given trace.
    """
    # first find where the maximum is
    n_samples = trace.shape[sample_axis]
    max_indces = jnp.median(jnp.argmax(trace, axis=sample_axis))
    to_roll = jnp.int32(n_samples // 2 - max_indces)

    # now generate a generic time array based on the number of samples
    time_axes = (
        jnp.tile(
            jnp.expand_dims(jnp.arange(n_samples), axis=0),
            reps=(arrival_times.shape[0], 1),
        )
        * delta_t
    )

    # now set the zero to be at the maximum and set the zero value to be the arrival time
    time_axes = time_axes - (time_axes[:, to_roll] + arrival_times)[:, None]

    return time_axes


def get_fluences(
    efield_traces: jax.typing.ArrayLike,
    delta_t: float,
    pol_axis: int = 0,
    sample_axis: int = 2,
) -> jax.Array:
    """
    Return the fluences from electric field traces.

    Parameters
    ----------
    efield_traces : jax.typing.ArrayLike
        the electric field traces
    delta_t : float
        the timing resolution in ns
    pol_axis : int, default=2
        the axis in which the polarisations are located.
    """
    eps_0 = constants.epsilon_0 * units.farad / units.m
    c_vacuum = constants.c * units.m / units.s
    return (
        eps_0
        * c_vacuum
        * jnp.sum(
            delta_t * jnp.linalg.norm(efield_traces, axis=pol_axis),
            axis=sample_axis - 1,
        )
    )
